module csr_matrix_mod
  use basics
  implicit none
  private
  public  csr_matrix, &
       matrix_initial, &
       matrix_allocate, &
       coo_to_csr_format, &
       matrix_deallocate, &
       matrix_multiply,&
       matrix_multiply_flops
  interface matrix_initial
     module procedure csr_matrix_initial
  end interface
  interface matrix_multiply
     module procedure csr_matrix_multiply
  end interface
  interface matrix_allocate
     module procedure csr_matrix_allocate
  end interface
  interface matrix_deallocate
     module procedure csr_matrix_deallocate
  end interface
  interface matrix_multiply_flops
     module procedure csr_matrix_multiply_flops
  end interface
  type csr_matrix
     integer (kind = myint) :: size
     integer (kind = myint) :: size_column
     integer (kind = myint), dimension (:), pointer :: ia,ja
     real (kind = myreal), dimension (:), pointer :: a
     logical :: structure_only
  end type csr_matrix

contains

  ! initialise the matrix to n rows 
  subroutine csr_matrix_initial(n,matrix,m,structure_only) 
    type (csr_matrix) :: matrix
    integer (kind = myint) :: i,ierr,n
    ! number of columns in the matrix
    integer (kind = myint), optional, intent (in) :: m
    ! whether this matrix is structure only
    logical, optional, intent (in) :: structure_only

    matrix%size = n
    if (present(m)) then
       matrix%size_column = m
    end if
    nullify(matrix%ja, matrix%a)

    allocate(matrix%ia(n+1),stat = ierr)

    if (ierr /= 0) stop "allocation error in csr_matrix_initial"
    !    default the matrix to be not structure only
    if (present(structure_only)) then
       matrix%structure_only = structure_only
    else
       matrix%structure_only = .false.
    end if


  end subroutine csr_matrix_initial

  subroutine csr_matrix_allocate(matrix,nz)
    ! allocate space for a matrix of nz zeros, also set the number
    ! of non-zeros to nz
    type (csr_matrix), intent (inout) :: matrix
    integer (kind = myint), intent (in) :: nz
    integer (kind = myint) :: ierr,nz1

    nz1 = nz
    if (nz1 <= 0) then
       write(*,*) "try to allocating ",nz," nonzeros in csr_matrix_allocate!"
       write(*,*) "set to 1"
       !       stop
       nz1 = 1
    end if
    allocate(matrix%ja(nz1),stat = ierr)
    if (ierr /= 0) stop "allocation error in csr_matrix_allocate"

    if (matrix%structure_only) return
    allocate(matrix%a(nz1),stat = ierr)
    if (ierr /= 0) stop "allocation error in csr_matrix_allocate"

  end subroutine csr_matrix_allocate

  subroutine csr_matrix_deallocate(matrix) 
    ! deallocate the space occupied by a csr formatted matrix
    type (csr_matrix), intent (inout) :: matrix

    integer (kind = myint) :: i,n,ierr


    deallocate(matrix%ja, stat = ierr)
    if (ierr > 0) write(*,*) &
         "deallocation error in csr_matrix_deallocate1"


    deallocate(matrix%ia, stat = ierr)
    if (ierr > 0) write(*,*) &
         "deallocation error in csr_matrix_deallocate3"

    if (matrix%structure_only) then
       return
    end if

    deallocate(matrix%a, stat = ierr)
    if (ierr > 0) write(*,*) &
         "deallocation error in csr_matrix_deallocate2"

  end subroutine csr_matrix_deallocate


  subroutine csr_matrix_multiply(matrix1,matrix2,result_matrix,m1) 
    ! multiply two matrices matrix1 and matrix2
    ! and put the resulting matrix in result_matrix
    type (csr_matrix), intent (in) :: matrix1,matrix2
    type (csr_matrix), intent (out) :: result_matrix
    ! size of the second dimension of matrix2
    integer (kind = myint), optional, intent (in) :: m1
    integer (kind = myint) :: n,nz,m
    real (kind = myreal), pointer, dimension(:) :: a,b,c
    integer (kind = myint), pointer, dimension(:) :: ia,ib,ic,ja,jb,jc

    n = matrix1%size
    if (present(m1)) then
       m = m1
    else
       m = matrix2%size_column
    end if


    call csr_matrix_initial(n,result_matrix,matrix2%size_column, &
         structure_only = .false.)

    ia => matrix1%ia; ib => matrix2%ia; ic => result_matrix%ia
    ja => matrix1%ja; jb => matrix2%ja; 
    a => matrix1%a; b => matrix2%a; 


    ! now find the size
    call matmul_size(n,m,ia,ja,ib,jb,nz)
    call csr_matrix_allocate(result_matrix,nz)
    jc => result_matrix%ja


    c => result_matrix%a
    call matmul_normal(n,m,ia,ja,a,ib,jb,b,ic,jc,c)

  end subroutine csr_matrix_multiply

  function csr_matrix_multiply_flops(n,m,ia,ja,ib,jb) result(flops)
    ! find the size of the new matrix which is a product of a1 and a2
    integer (kind = myint), intent (in) :: n,m
    integer (kind = myint), intent (in), dimension (*) :: ia,ja,ib,jb
    ! flops: the number of flops during the multiplication 
    ! (= matmul_size + matmul_normal)
    integer (kind = myint) :: flops

    integer (kind = myint) :: nz
    integer (kind = myint), dimension (m) :: mask
    integer (kind = myint) :: i,j,k,neigh,icol_add

    ! initialise the mask array which is an array that has 
    ! value i if column index already exist in row i of the new matrix
    mask = 0
    nz = 0
    flops = 0
    do i=1,n
       do j = ia(i),ia(i+1)-1
          neigh = ja(j)
          do k = ib(neigh),ib(neigh+1)-1
             icol_add = jb(k)
             flops = flops + 2
             if (mask(icol_add) /= i) then
                nz = nz + 1
                mask(icol_add) = i     ! add mask
             end if
          end do
       end do
    end do
    flops = flops + nz
  end function csr_matrix_multiply_flops




  subroutine matmul_size(n,m,ia1,ja1,ia2,ja2,nz)
    ! find the size of the new matrix which is a product of a1 and a2
    integer (kind = myint), intent (in) :: n,m
    integer (kind = myint), intent (in), dimension (*) :: ia1,ja1,ia2,ja2
    integer (kind = myint), intent (inout) :: nz
    integer (kind = myint), dimension (m) :: mask
    integer (kind = myint) :: i,j,k,neigh,icol_add

    ! initialise the mask array which is an array that has 
    ! value i if column index already exist in row i of the new matrix
    mask = 0
    nz = 0
    do i=1,n
       do j = ia1(i),ia1(i+1)-1
          neigh = ja1(j)
          do k = ia2(neigh),ia2(neigh+1)-1
             icol_add = ja2(k)
             if (mask(icol_add) /= i) then
                nz = nz + 1
                mask(icol_add) = i     ! add mask
             end if
          end do
       end do
    end do
  end subroutine matmul_size


  subroutine matmul_normal(n,m,ia,ja,a,ib,jb,b,ic,jc,c)
    ! C = A*B
    integer (kind = myint), intent (in) :: n,m
    integer (kind = myint), intent (in) :: ia(*),ib(*),ja(*),jb(*)
    integer (kind = myint), intent (out) :: ic(*),jc(*)
    real (kind = myreal), intent (in) :: a(*),b(*)
    real (kind = myreal), intent (out) :: c(*)
    integer (kind = myint) :: mask(m),nz,i,j,k,icol,icol_add,neigh
    real (kind = myreal) :: aij

    ! initialise the mask array which is an array that has non-zero value if the 
    ! column index already exist, in which case
    ! the value is the index of that column 
    mask = 0
    nz = 0
    ic(1) = 1
    do i=1,n
       do j = ia(i),ia(i+1)-1
          aij = a(j)
          neigh = ja(j)
          do k = ib(neigh),ib(neigh+1)-1
             icol_add = jb(k)
             icol = mask(icol_add)
             if (icol == 0) then
                nz = nz + 1
                jc(nz) = icol_add
                c(nz) =  aij*b(k)
                mask(icol_add) = nz     ! add mask
             else
                c(icol) = c(icol) + aij*b(k)
             end if
          end do
       end do
       mask(jc(ic(i):nz)) = 0   ! done this row i, so set mask to zero again
       ic(i+1) = nz+1
    end do
  end subroutine matmul_normal

  subroutine coo_to_csr_format(n,nz,irn,jcn,val,matrix,structure_only,ncol) 
    ! convert irn,jcn,val to CSR matrix format
    type (csr_matrix) :: matrix
    integer (kind=myint), intent (in) :: irn(*)
    integer (kind=myint), intent (in) :: jcn(*)
    real (kind = myreal), intent (in) :: val(*)
    integer (kind=myint), intent (in) :: n, nz
    logical, intent (in), optional :: structure_only
    integer (kind=myint), intent (in), optional :: ncol

    integer (kind=myint) :: size_column
    integer (kind=myint) :: ierr
    integer (kind=myint) :: i
    integer (kind = myint) :: nrow(n),ii,jj,iia(n+1)
    integer (kind = myint), pointer, dimension (:) :: ia

    if (present(ncol)) then
       size_column = ncol
    else
       size_column = maxval(jcn(1:nz))
    end if
    if (present(structure_only)) then
       call csr_matrix_initial(n,matrix,size_column,structure_only = structure_only)
    else
       call csr_matrix_initial(n,matrix,size_column,structure_only = .false.)
    end if
    ia => matrix%ia

    nrow = 0
    do i = 1, nz
       nrow(irn(i)) = nrow(irn(i)) + 1
    end do
    ia (1) = 1
    do i = 1, n
       ia(i+1) = ia(i) + nrow(i)
    end do

    call csr_matrix_allocate(matrix,nz)

    iia(1:n+1) = ia(1:n+1)
    do i = 1, nz
       ii = irn(i)
       jj = jcn (i)
       matrix%ja(iia(ii)) = jj
       if (.not.matrix%structure_only) matrix%a(iia(ii)) = val(i)
       iia(ii) = iia(ii) + 1
    end do

  end subroutine coo_to_csr_format


end module csr_matrix_mod
